import sys
from .result import Result
import inspect
import traceback


def check_scatter_chart(option, value):
    if option == "hist_color":
        if value not in ["r", "g", "b"]:
            raise Exception('"hist_color" takes only values r or g or b')


def check(function, option, value):
    if function == "scatterchart":
        check_scatter_chart(option, value)


def check_params(params, options, funcname):
    for k in params:
        if k not in options:
            keys_list = ", ".join(list(options.keys()))
            sys.exit(
                f'Option "{k}" for {funcname} is not valid, available options: '
                + keys_list
            )


def check_color(col):
    if isinstance(col, int) or isinstance(col, float):
        return Result.Ok((col, col, col))
    elif isinstance(col, list) or isinstance(col, tuple):
        if not len(col) in [1, 3]:
            return Result.Fail("Fill takes only 1 or 3 parameters when using a list")
        if len(col) == 1:
            return Result.Ok((col[0], col[0], col[0]))
        else:
            return Result.Ok((col[0], col[1], col[2]))
    else:
        return Result.Fail("Fill should be of type int, float, list or tuple")


def check_value(val, type_, value_name, range_=None):
    if not isinstance(value_name, str):
        return Result.Fail("Value name should be of type string")
    else:
        if not isinstance(type_, list):
            if not isinstance(val, type_):
                return Result.Fail(
                    f'Value "{val}" for {value_name} should be of type {type_}'
                )
            if type_ in [int, float]:
                if type(range_) in [list, tuple]:
                    if len(range_) == 2:
                        if not range_[0] < val < range_[1]:
                            return Result.Fail(
                                f"Value of {value_name} should be between {range_[0]} and {range_[1]}"
                            )
                    elif len(range_) == 0:
                        pass
                    else:
                        return Result.Fail("Range takes [n1, n2] or []")
                else:
                    return Result.Fail("Range should be of type list or tuple")
        elif isinstance(type_, list):
            results = []
            for t in type_:
                if not isinstance(val, t):
                    results.append(0)
                else:
                    results.append(1)
            if not any(results):
                return Result.Fail(
                    f'Value "{val}" for {value_name} should be of type {type_}'
                )

    return Result.Ok(val)


def verify_func_param(method, param_types, locals_):
    try:
        param_names = inspect.getfullargspec(method)[0]
        for param in param_names:
            if param not in ["self"]:
                verif = check_value(
                    locals_[param],
                    param_types[param][0],
                    param,
                    range_=param_types[param][1],
                )
                if not verif.success:
                    raise Exception(verif.error)
    except Exception as e:
        filename = traceback.extract_stack()[0].filename
        line = traceback.extract_stack()[0].line
        lineno = traceback.extract_stack()[0].lineno
        print(
            "Hooman",
            type(e).__name__,  # TypeError
            "in file",
            filename,  # /tmp/example.py
            "at line",
            lineno,  # 2
        )
        print(">>>", line)
        print(str(e))
        sys.exit()


def verify_color(colors=[]):
    for color in colors:
        try:
            verif = check_color(color)
            if not verif.success:
                raise Exception(verif.error)
        except Exception as e:
            filename = traceback.extract_stack()[0].filename
            line = traceback.extract_stack()[0].line
            lineno = traceback.extract_stack()[0].lineno
            print(
                "Hooman",
                type(e).__name__,  # TypeError
                "in file",
                filename,  # /tmp/example.py
                "at line",
                lineno,  # 2
            )
            print(">>>", line)
            print(str(e))
            sys.exit()



import copy

from .check import check_params


def barchart(hapi, x, y, w, h, params):
    options = {
        "ticks_y": 10,
        "tick_size": 5,
        "range_y": [0, 100],
        "data": {"a": 10, "b": 20},
        "bin_color": (255, 99, 97),
        "line_color": (200, 200, 200),
        "text_color": (100, 100, 100),
        "mouse_line": False,
    }
    check_params(params, options, "bar chart")
    options.update(params)
    hapi.stroke_size(2)
    hapi.stroke(options["line_color"])
    hapi.line(x, y, x, y + h)
    hapi.line(x, y + h, x + w, y + h)
    hapi.fill(options["text_color"])
    y_top_val = options["range_y"][1]
    for t in range(options["ticks_y"]):
        hapi.line(
            x,
            y + (t * (h // options["ticks_y"])),
            x - options["tick_size"],
            y + (t * (h // options["ticks_y"])),
        )
        hapi.text(y_top_val, x - 20, y + (t * (h // options["ticks_y"])) - 10)
        y_top_val -= (options["range_y"][1] - options["range_y"][0]) // options[
            "ticks_y"
        ]

    num_data = len(options["data"].keys())
    divs = 1 + 1 + (num_data * 2) + (num_data - 1)
    space_unit = w // divs
    current_x = x + space_unit
    for i, d in enumerate(options["data"]):
        hapi.fill(options["bin_color"])
        bin_height = hapi.constrain(
            options["data"][d], options["range_y"][0], options["range_y"][1], 0, h
        )
        hapi.rotate(0)
        hapi.rect(current_x, y + h - bin_height, 2 * space_unit, bin_height)
        hapi.fill(hapi.color["black"])
        hapi.push_matrix()
        hapi.rotate(270)
        hapi.text(d, current_x + space_unit, y + h + 5)
        hapi.pop_matrix()
        current_x += 3 * space_unit

    if options["mouse_line"]:
        hapi.stroke_size(2)
        hapi.stroke(hapi.color["red"])
        l_height = hapi.mouseY()
        if hapi.mouseY() < y:
            l_height = y
        elif hapi.mouseY() > y + h:
            l_height = y + h
        hapi.line(x, l_height, x + w, l_height)


def linechart(hapi, x, y, w, h, params):
    options = {
        "ticks_y": 10,
        "ticks_x": 10,
        "tick_size": 5,
        "range_y": [0, 100],
        "range_x": [0, 100],
        "lines": [
            {
                "label": "---",
                "color": (255, 0, 0),
                "data": [[1, 1]],
                "values_window": 200,
            }
        ],
        "labels": ["apple", "", "", "tree"],
        "line_color": (200, 200, 200),
        "text_color": (100, 100, 100),
        "mouse_line": False,
        "mouse_line_color": (255, 0, 0),
        "graph_color": (0, 0, 0),
        "show_axes": True,
        "show_ticks_x": True,
        "show_ticks_y": True,
        "x_axis_label": "x_axis_label",
        "y_axis_label": "y_axis_label",
        "plot_background": True,
        "plot_grid": True,
        "plot_background_color": (234, 234, 242),
        "plot_grid_color": 255,
    }
    check_params(params, options, "line chart")
    options.update(params)
    hapi.stroke_size(2)

    # --- plot background ---

    if options["plot_background"]:

        hapi.fill(options["plot_background_color"])
        hapi.rect(x, y, w, h)

    # --- plot background grid ---

    if options["plot_grid"]:

        hapi.stroke(options["plot_grid_color"])

        # --- vertical
        bg_grid_x1 = x
        bg_grid_y1 = y
        bg_grid_x2 = x
        bg_grid_y2 = y + h
        line_num = 1
        while bg_grid_x1 <= x + w:
            if 1 < line_num < options["ticks_x"] + 1:
                hapi.line(bg_grid_x1, bg_grid_y1, bg_grid_x2, bg_grid_y2)
            bg_grid_x1 += w // options["ticks_x"]
            bg_grid_x2 += w // options["ticks_x"]
            line_num += 1

        # --- horizontal
        bg_grid_x1 = x
        bg_grid_y1 = y
        bg_grid_x2 = x + w
        bg_grid_y2 = y
        line_num = 1
        while bg_grid_y1 <= y + h:
            if 1 < line_num < options["ticks_y"] + 1:
                hapi.line(bg_grid_x1, bg_grid_y1, bg_grid_x2, bg_grid_y2)
            bg_grid_y1 += h // options["ticks_y"]
            bg_grid_y2 += h // options["ticks_y"]
            line_num += 1

    # --- axes ---

    if options["show_axes"]:
        hapi.stroke(options["line_color"])
        hapi.line(x, y, x, y + h)
        hapi.line(x, y + h, x + w, y + h)

    hapi.fill(options["text_color"])

    # --- axes labels

    hapi.push_matrix()
    hapi.rotate(90)
    hapi.text(
        options["x_axis_label"],
        x - 35,
        y + (h - (hapi._font.size(options["x_axis_label"])[0]) // 2),
    )
    hapi.pop_matrix()

    hapi.text(
        options["y_axis_label"],
        x + (w - (hapi._font.size(options["y_axis_label"])[0]) // 2),
        y + h + 35,
    )

    y_top_val = options["range_y"][1]
    for t in range(options["ticks_y"]):
        if options["show_ticks_y"]:
            hapi.line(
                x,
                y + (t * (h // options["ticks_y"])),
                x - options["tick_size"],
                y + (t * (h // options["ticks_y"])),
            )
        hapi.text(y_top_val, x - 20, y + (t * (h // options["ticks_y"])) - 10)
        y_top_val -= (options["range_y"][1] - options["range_y"][0]) // options[
            "ticks_y"
        ]

    x_val = options["range_x"][0]
    for t in range(options["ticks_x"] + 1):
        if options["show_ticks_x"]:
            hapi.line(
                x + (t * (w // options["ticks_x"])),
                y + h,
                x + (t * (w // options["ticks_x"])),
                y + h + options["tick_size"],
            )
        hapi.push_matrix()
        hapi.rotate(270)
        hapi.text(x_val, x + (t * (w // options["ticks_x"])), y + h + 5)
        hapi.pop_matrix()
        x_val += (options["range_x"][1] - options["range_x"][0]) // options["ticks_x"]

    for l_index, line in enumerate(copy.deepcopy(options["lines"])):
        for i, d in enumerate(copy.deepcopy(line["data"])):

            try:
                x1 = hapi.constrain(
                    d[0], options["range_x"][0], options["range_x"][1], x, x + w
                )
                y1 = (
                    y
                    + h
                    - hapi.constrain(
                        d[1], options["range_y"][0], options["range_y"][1], y, y + h
                    )
                    + y
                )

                x2 = hapi.constrain(
                    line["data"][i + 1][0],
                    options["range_x"][0],
                    options["range_x"][1],
                    x,
                    x + w,
                )
                y2 = (
                    y
                    + h
                    - hapi.constrain(
                        line["data"][i + 1][1],
                        options["range_y"][0],
                        options["range_y"][1],
                        y,
                        y + h,
                    )
                    + y
                )

                # 60 is arbitrary value
                # hapi.fill(hapi.color['yellow'])
                # print(x1, y1)
                hapi.stroke_size(2)
                hapi.stroke(line["color"])
                hapi.line(x1, y1, x2, y2)
                # hapi.line(x1, y1, x2, y2)
            except IndexError:
                pass

            # if len(line["data"]) > line["values_window"]:
            # # options["lines"][l_index]["data"].pop(0)

            #     del options["lines"][l_index]['data'][0]

    if options["mouse_line"]:
        hapi.stroke_size(2)
        hapi.stroke(options["mouse_line_color"])
        limit_y = hapi.mouseY()
        limit_x = hapi.mouseX()
        if limit_y < y:
            limit_y = y
        elif limit_y > y + h:
            limit_y = y + h
        if limit_x < x:
            limit_x = x
        elif limit_x > x + w:
            limit_x = x + w

        hapi.line(x, limit_y, x + w, limit_y)
        hapi.line(limit_x, y, limit_x, y + h)
        # tex = "{} {}".format(limit_x, limit_y)
        # hapi.text(tex, 100, 100)

    # legend

    legend_x_start = x + w + 10
    legend_y_start = y
    legend_y = legend_y_start

    for line in options["lines"]:
        hapi.fill(line["color"])
        hapi.ellipse(legend_x_start, legend_y, 10, 10)
        hapi.font_size(15)
        hapi.text(line["label"], legend_x_start + 15, legend_y)
        legend_y += 10


def piechart(hapi, x, y, radius, data, start_rad=0):
    first_run = True
    total = 0
    if first_run:
        for d in data:
            total += d[1]
        first_run = False

    previous_a = 0
    for i, d in enumerate(data):
        hapi.fill(d[2])
        end_a = previous_a + (d[1] / total) * 360
        hapi.fill_arc(x, y, radius, previous_a, end_a, start_rad=start_rad)
        previous_a = end_a


def scatterchart(hapi, x, y, w, h, params):
    options = {
        "ticks_y": 10,
        "ticks_x": 10,
        "tick_size": 5,
        "range_y": [0, 100],
        "range_x": [0, 100],
        "data": {
            "carat": [0.23, 0.21, 0.23, 0.29, 0.31, 0.24, 0.24, 0.26, 0.22, 0.23],
            "cut": [
                "Ideal",
                "Premium",
                "Good",
                "Premium",
                "Good",
                "Very Good",
                "Very Good",
                "Very Good",
                "Fair",
            ],
            "color": ["E", "E", "E", "I", "J", "J", "I", "H", "E", "H"],
            "clarity": [
                "SI2",
                "SI1",
                "VS1",
                "VS2",
                "SI2",
                "VVS2",
                "VVS1",
                "SI1",
                "VS2",
                "VS1",
            ],
            "depth": [61.5, 59.8, 56.9, 62.4, 63.3, 62.8, 62.3, 61.9, 65.1, 59.4],
            "table": [55, 61, 65, 58, 58, 57, 57, 55, 61, 61],
            "price": [326, 326, 327, 334, 335, 336, 336, 337, 337, 338],
            "x": [3.95, 3.89, 4.05, 4.2, 4.34, 3.94, 3.95, 4.07, 3.87, 4],
            "y": [3.98, 3.84, 4.07, 4.23, 4.35, 3.96, 3.98, 4.11, 3.78, 4.05],
            "z": [2.43, 2.31, 2.31, 2.63, 2.75, 2.48, 2.47, 2.53, 2.49, 2.39],
        },
        "hue": None,
        "hue_order": [],
        "size": None,
        "text_color": (100, 100, 100),
        "mouse_line": False,
        "mouse_line_color": (255, 0, 0),
        "graph_color": (0, 0, 0),
        "show_axes": True,
        "show_ticks_x": True,
        "show_ticks_y": True,
        "x": "price",
        "y": "carat",
        "plot_background": True,
        "plot_grid": True,
        "plot_background_color": (234, 234, 242),
        "plot_grid_color": 255,
        "line_color": 200,
        "strong_color": (107, 107, 255),
        "light_color": (235, 235, 255),
        "type": "normal",
        "kind": "rect",
        "hist_color": "b",
        "hist_color_invert": False,
    }

    check_params(params, options, "scatter chart")
    options.update(params)

    hapi.stroke_size(2)
    hapi.font_size(15)

    # --- plot background ---

    if options["plot_background"]:

        hapi.fill(options["plot_background_color"])
        hapi.rect(x, y, w, h)

    # --- mouse line ---

    if options["mouse_line"]:
        hapi.stroke_size(2)
        hapi.stroke(options["mouse_line_color"])
        limit_y = hapi.mouseY()
        limit_x = hapi.mouseX()
        if limit_y < y:
            limit_y = y
        elif limit_y > y + h:
            limit_y = y + h
        if limit_x < x:
            limit_x = x
        elif limit_x > x + w:
            limit_x = x + w

        hapi.line(x, limit_y, x + w, limit_y)
        hapi.line(limit_x, y, limit_x, y + h)

    # --- scatter ---

    xvals = options["data"][options["x"]]
    yvals = options["data"][options["y"]]
    # print(xvals[:10], yvals[:10])
    # hapi.set_alpha(100)

    if options["hue"]:
        strong_blue = options["strong_color"]
        light_blue = options["light_color"]
        strong_blue_hls = hapi.rgb_to_hls(*strong_blue)
        light_blue_hls = hapi.rgb_to_hls(*light_blue)
        l_strong = strong_blue_hls[1]
        l_light = light_blue_hls[1]

        delta_l = l_strong - l_light
        steps = len(options["hue_order"])
        delta_change = delta_l / steps

        c_map = {}

        color_l = l_strong

        legend_y = y + 10
        hapi.fill(0)
        hapi.text(options["hue"], x + w + 10, y)
        for c in options["hue_order"]:
            color_hue = hapi.hls_to_rgb(strong_blue_hls[0], color_l, strong_blue_hls[2])
            c_map[c] = color_hue
            color_l += delta_change

            # --- legend

            # --- --- hue

            hapi.fill(color_hue)
            hapi.ellipse(x + w + 10, legend_y, 10, 10)
            hapi.text(c, x + w + 10 + 15, legend_y)
            legend_y += 10

    hue = options["hue"]

    # --- --- size

    if options["size"]:
        min_size = min(options["data"][options["size"]])
        max_size = max(options["data"][options["size"]])

        hapi.fill(0)
        legend_y += 10
        hapi.text(options["size"], x + w + 10, legend_y)
        legend_y += 10

        d_size = (max_size - min_size) / 5
        start_size = min_size
        ellipse_size = 0
        for i in range(5):
            ellipse_size = i + 1
            hapi.ellipse(x + w + 10, legend_y, ellipse_size, ellipse_size)
            hapi.text(round(start_size, 2), x + w + 15, legend_y)
            start_size += d_size
            legend_y += 10

        ellipse_size += 1
        hapi.ellipse(x + w + 10, legend_y, ellipse_size, ellipse_size)
        hapi.text(round(start_size, 2), x + w + 15, legend_y)

    # /--- legend

    # --- actual plotting

    if options["type"] == "normal":
        for i, vx in enumerate(xvals):
            if options["hue"]:
                current_hue = options["data"][options["hue"]][i]
                current_color = c_map[current_hue]
            else:
                current_color = options["graph_color"]

            hapi.fill(current_color)

            ex = hapi.constrain(
                vx,
                options["range_x"][0],
                options["range_x"][1],
                x,
                x + w,
            )
            ey = (
                y
                + h
                - hapi.constrain(
                    yvals[i],
                    options["range_y"][0],
                    options["range_y"][1],
                    y,
                    y + h,
                )
                + y
            )

            if options["size"]:
                size = options["data"][options["size"]][i]
                ellipse_size = (size / max_size) * 5
            else:
                ellipse_size = 5
            hapi.ellipse(ex, ey, ellipse_size, ellipse_size)
    elif options["type"] == "hist":

        range_x = (
            max(options["data"][options["x"]]) - min(options["data"][options["x"]])
        ) // options["ticks_x"]
        range_y = (
            max(options["data"][options["y"]]) - min(options["data"][options["y"]])
        ) // options["ticks_y"]

        m = {}

        # for xi in range(options['ticks_x']+1):
        #     for yi in range(options['ticks_y']+1):
        #         m[str(xi*range_x)+'-'+str(yi*range_y)] = 0

        total = 0
        for x_point, y_point in zip(
            options["data"][options["x"]], options["data"][options["y"]]
        ):
            xi = x_point / (
                (options["range_x"][1] - options["range_x"][0]) / options["ticks_x"]
            )
            yi = y_point / (
                (options["range_y"][1] - options["range_y"][0]) / options["ticks_y"]
            )

            yi = options["ticks_y"] - yi

            xi = int(xi)
            yi = int(yi)

            if (xi, yi) not in m:
                m[(xi, yi)] = 1
                total += 1
            else:
                m[(xi, yi)] += 1
                total += 1

        li = [[k, m[k]] for k in m]

        sorted_li = sorted(li, key=lambda l: l[1], reverse=True)

        rect_w = w // options["ticks_x"]
        rect_h = h // options["ticks_y"]

        max_key = max(m, key=m.get)
        max_val = m[max_key]
        for xi in range(options["ticks_x"]):
            for yi in range(options["ticks_y"]):
                if (xi, yi) in m:

                    if options["hist_color_invert"]:
                        col = (m[(xi, yi)] / max_val) * 255
                    else:
                        col = (((max_val - m[(xi, yi)]) / max_val)) * 255

                    check("scatterchart", "hist_color", options["hist_color"])
                    if options["hist_color"] == "r":
                        col_rect = (int(col), 0, 0)
                    if options["hist_color"] == "g":
                        col_rect = (0, int(col), 0)
                    if options["hist_color"] == "b":
                        col_rect = (0, 0, int(col))
                    hapi.fill(col_rect)
                else:
                    hapi.fill(255)

                if options["kind"] == "rect":
                    hapi.rect(x + xi * rect_w, y + yi * rect_h, rect_w, rect_h)
                if options["kind"] == "round":
                    hapi.ellipse(x + xi * rect_w, y + yi * rect_h, rect_w, rect_h)
                elif options["kind"] == "hex":
                    err = 1.8 * rect_w
                    if yi % 2 == 0:
                        hapi.regular_polygon(
                            x + xi * rect_w, y + yi * rect_h, rect_w - err, rect_h, 6
                        )
                    else:
                        hapi.regular_polygon(
                            x + xi * rect_w + (err // 4),
                            y + yi * rect_h,
                            rect_w - err,
                            rect_h,
                            6,
                        )

    # --- plot grid ---

    hapi.stroke_size(1)
    if options["plot_grid"]:

        hapi.stroke(options["plot_grid_color"])

        # --- vertical
        bg_grid_x1 = x
        bg_grid_y1 = y
        bg_grid_x2 = x
        bg_grid_y2 = y + h
        line_num = 1
        while bg_grid_x1 <= x + w:
            if 1 < line_num < options["ticks_x"] + 1:
                hapi.line(bg_grid_x1, bg_grid_y1, bg_grid_x2, bg_grid_y2)
            bg_grid_x1 += w // options["ticks_x"]
            bg_grid_x2 += w // options["ticks_x"]
            line_num += 1

        # --- horizontal
        bg_grid_x1 = x
        bg_grid_y1 = y
        bg_grid_x2 = x + w
        bg_grid_y2 = y
        line_num = 1
        while bg_grid_y1 <= y + h:
            if 1 < line_num < options["ticks_y"] + 1:
                hapi.line(bg_grid_x1, bg_grid_y1, bg_grid_x2, bg_grid_y2)
            bg_grid_y1 += h // options["ticks_y"]
            bg_grid_y2 += h // options["ticks_y"]
            line_num += 1

    # --- axes ---

    if options["show_axes"]:
        hapi.stroke(options["line_color"])
        hapi.line(x, y, x, y + h)
        hapi.line(x, y + h, x + w, y + h)

    hapi.fill(options["text_color"])

    # --- axes labels

    hapi.push_matrix()
    hapi.rotate(90)
    hapi.text(options["y"], x - 35, y + (h - (hapi._font.size(options["y"])[0]) // 2))
    hapi.pop_matrix()

    hapi.text(
        options["x"], x + (w - (hapi._font.size(options["x"])[0]) // 2), y + h + 35
    )

    # --- axes

    y_top_val = options["range_y"][1]
    for t in range(options["ticks_y"]):
        if options["show_ticks_y"]:
            hapi.line(
                x,
                y + (t * (h // options["ticks_y"])),
                x - options["tick_size"],
                y + (t * (h // options["ticks_y"])),
            )
        hapi.text(y_top_val, x - 20, y + (t * (h // options["ticks_y"])) - 10)
        y_top_val -= (options["range_y"][1] - options["range_y"][0]) // options[
            "ticks_y"
        ]

    x_val = options["range_x"][0]
    for t in range(options["ticks_x"] + 1):
        if options["show_ticks_x"]:
            hapi.line(
                x + (t * (w // options["ticks_x"])),
                y + h,
                x + (t * (w // options["ticks_x"])),
                y + h + options["tick_size"],
            )
        hapi.push_matrix()
        hapi.rotate(270)
        hapi.text(round(x_val, 2), x + (t * (w // options["ticks_x"])), y + h + 5)
        hapi.pop_matrix()
        x_val += (options["range_x"][1] - options["range_x"][0]) / options["ticks_x"]

        # print(x_val, ((options["range_x"][1] - options["range_x"][0]) / options["ticks_x"]))



class Result:
    # Thanks Aaron Luna
    """Represent the outcome of an operation."""

    def __init__(self, success, value, error):
        """Represent the outcome of an operation."""
        self.success = success
        self.error = error
        self.value = value

    def __str__(self):
        """Informal string representation of a result."""
        if self.success:
            return "[Success]"
        else:
            return f"[Failure] {self.error}"

    def __repr__(self):
        """Official string representation of a result."""
        if self.success:
            return f"<Result success={self.success}>"
        else:
            return f'<Result success={self.success}, message="{self.error}">'

    @property
    def failure(self):
        """Flag that indicates if the operation failed."""
        return not self.success

    def on_success(self, func, *args, **kwargs):
        """Pass result of successful operation (if any) to subsequent function."""
        if self.failure:
            return self
        if self.value:
            return func(self.value, *args, **kwargs)
        return func(*args, **kwargs)

    def on_failure(self, func, *args, **kwargs):
        """Pass error message from failed operation to subsequent function."""
        if self.success:
            return self.value if self.value else None
        if self.error:
            return func(self.error, *args, **kwargs)
        return func(*args, **kwargs)

    def on_both(self, func, *args, **kwargs):
        """Pass result (either succeeded/failed) to subsequent function."""
        if self.value:
            return func(self.value, *args, **kwargs)
        return func(*args, **kwargs)

    @staticmethod
    def Fail(error_message):
        """Create a Result object for a failed operation."""
        return Result(False, value=None, error=error_message)

    @staticmethod
    def Ok(value=None):
        """Create a Result object for a successful operation."""
        return Result(True, value=value, error=None)

    @staticmethod
    def Combine(results):
        """Return a Result object based on the outcome of a list of Results."""
        if all(result.success for result in results):
            return Result.Ok()
        errors = [result.error for result in results if result.failure]
        return Result.Fail("\n".join(errors))



class SVG:
    @classmethod
    def tag(
        cls, tag_name: str, content: str = "", attributes: dict = {}, self_close=False
    ) -> str:
        attribs = []
        for k in attributes:
            if attributes[k]:
                attribs.append(f'{k}="{str(attributes[k])}"')

        attribs_str = " ".join(attribs)

        if not self_close:
            return f"<{tag_name} {attribs_str}>{content}</{tag_name}>"
        else:
            return f"<{tag_name} {attribs_str}/>"

    @classmethod
    def save(cls, commands, path, width, height):
        commands_str = "\n".join(commands)
        string = f"""
<svg version="1.1"
     width="{width}" height="{height}"
     xmlns="http://www.w3.org/2000/svg">

{commands_str}

</svg>
"""

        with open(path, "w+") as f:
            f.write(string)

        if len(commands) == 0:
            print("warning! SVG mode not enabled!")
        print("saved svg to", path)



import sys
import os 
import shutil
import uuid

import pygame
import pygame.gfxdraw

from math import pi
from math import cos
from math import sin
from math import sqrt
import datetime

from .ui import Button
from .ui import Slider
from .ui import TextBox
from .ui import slider_with_text
from .ui import Scroll

from .shapes import star
from .shapes import alpha_ellipse
from .shapes import curve_rect
from .shapes import arrow
from .shapes import heart
from .shapes import regular_polygon
from .shapes import supershape
from .shapes import smooth_star
from .shapes import flowing_star
from .shapes import oil_drop
from .shapes import ellipse
from .shapes import cross_hair
from .shapes import gradient_rect

from .formula import constrain
from .formula import round_to_num
from .formula import distance
from .formula import rgb_to_hls
from .formula import hls_to_rgb

from .time import Timer

from .charts import barchart
from .charts import linechart
from .charts import piechart
from .charts import scatterchart

from .svg import SVG

from .check import check_color
from .check import verify_color
from .check import check_value
from .check import verify_func_param


class Hooman:
    def __init__(self, WIDTH, HEIGHT, svg=False):
        pygame.init()
        self.WIDTH = WIDTH
        self.HEIGHT = HEIGHT
        self.PI = pi
        self.center_x = WIDTH // 2
        self.center_y = HEIGHT // 2
        self.sin = sin
        self.cos = cos
        self.constrain = constrain
        self.sqrt = sqrt
        self.round_to = round_to_num
        self.dist = distance

        self.colors = {
            "red": (255, 0, 0),
            "green": (0, 255, 0),
            "blue": (0, 0, 255),
            "black": (0, 0, 0),
            "white": (255, 255, 255),
            "yellow": (255, 255, 0),
            "grey": (100, 100, 100),
            "light_grey": (200, 200, 200),
        }
        self.colours = self.colors
        self.color = self.colors
        self.colour = self.colors

        self.set_caption = pygame.display.set_caption
        self.screen = pygame.display.set_mode([WIDTH, HEIGHT])
        self.is_running = True
        self.bg_col = None
        self.set_caption("hooman window")

        self._rotation = 0
        self._alpha = 255
        self._fill = (255, 255, 255)
        self._stroke = (255, 255, 255)
        self._stroke_weight = 0
        self._font_name = "freesansbold.ttf"
        self._font_size = 32
        self._font = pygame.font.Font(self._font_name, self._font_size)
        self._matrix_stack = []

        self.sysfont = "comicsansms"
        self._font_size = 10
        self.pygame = pygame
        self.mouse_test_x = 0
        self.clock = pygame.time.Clock()
        self._all_widgets = []
        self._polygon_coords = []

        self._star = star
        self._alpha_ellipse = alpha_ellipse
        self._curve_rect = curve_rect
        self._arrow = arrow
        self._heart = heart
        self._reg_poly = regular_polygon
        self._supershape = supershape
        self._smooth_star = smooth_star
        self._flowing_star = flowing_star
        self._oil_drop = oil_drop
        self._cross_hair = cross_hair
        self._gradient = gradient_rect

        self._timers = []

        self._barchart = barchart
        self._linechart = linechart
        self._piechart = piechart
        self._scatterchart = scatterchart

        self.hls_to_rgb = hls_to_rgb
        self.rgb_to_hls = rgb_to_hls

        self._svg = svg
        self._svg_commands = []

        self.session_id = uuid.uuid1()
        self._session_vars = {}

    #
    # colors
    #

    def fill(self, col):
        """The color to fill drawn shapes with"""
        verify_color([col])
        self._fill = check_color(col).value

    def stroke(self, col):
        """The color to draw lines/strokes with"""
        verify_color([col])
        self._stroke = check_color(col).value

    def background(self, col):
        """Fill the screen with a color"""

        verify_color([col])
        self.screen.fill(check_color(col).value)

        # check_col = check_

    def gradient(self, w, h, start_col, end_col, direction=0):
        """returns a pygame.Surface with a gradient between 2 colors"""
        param_types = {
            "w": [int, []],
            "h": [int, []],
            "direction": [int, []],
        }
        verify_color([start_col, end_col])
        verify_func_param(self.gradient, param_types, locals())
        return self._gradient(
            w, h, check_color(start_col).value, check_color(end_col).value, direction
        )

    def set_background(self, col):
        """this calls hapi.background every frame with the given color"""

        verify_color([col])
        self.bg_col = check_color(col).value

    def stroke_size(self, weight):
        """The thickness of drawn lines"""
        param_types = {
            "weight": [int, []],
        }
        verify_func_param(self.stroke_size, param_types, locals())

        self._stroke_weight = weight

    def set_alpha(self, alpha):
        """sets the alpha to """
        self._alpha = alpha

    #
    # size
    #

    def no_stroke(self):
        self._stroke_weight = 0

    def font_size(self, font_size):
        param_types = {
            "font_size": [int, []],
        }

        verify_func_param(self.font_size, param_types, locals())

        self._font_size = font_size

    #
    # transforms
    #

    def rotate(self, angle):
        param_types = {
            "angle": [int, []],
        }
        verify_func_param(self.rotate, param_types, locals())
        self._rotation = angle % 360

    def push_matrix(self):
        self._matrix_stack.append(self._rotation)

    def pop_matrix(self):
        self._rotation = self._matrix_stack.pop()

    #
    # shapes
    #

    def ellipse(self, x, y, width, height):
        param_types = {
            "x": [[int, float], []],
            "y": [[int, float], []],
            "width": [[int, float], []],
            "height": [[int, float], []],
        }
        verify_func_param(self.ellipse, param_types, locals())

        pygame.draw.ellipse(self.screen, self._fill, (x, y, width, height))

        if self._svg:
            attributes = {
                "cx": x,
                "cy": y,
                "rx": width / 2,
                "ry": height / 2,
                "fill": f"rgb({self._fill[0]},{self._fill[1]},{self._fill[2]})",
                "stroke": f"rgb({self._stroke[0]},{self._stroke[1]},{self._stroke[2]})",
                "stroke-width": self._stroke_weight,
                "transform": f"rotate({self._rotation}, {x}, {y})",  # "translate(30,40) rotate(45)'
            }
            svg_element = SVG.tag("ellipse", attributes=attributes, self_close=True)
            self._svg_commands.append(svg_element)

    def rect(self, x, y, width, height):
        param_types = {
            "x": [[int, float], []],
            "y": [[int, float], []],
            "width": [[int, float], []],
            "height": [[int, float], []],
        }
        verify_func_param(self.rect, param_types, locals())
        if self._rotation % 360 == 0:
            pygame.draw.rect(self.screen, self._fill, (x, y, width, height))

            if self._svg:
                attributes = {
                    "x": x,
                    "y": y,
                    "width": width,
                    "height": height,
                    "fill": f"rgb({self._fill[0]},{self._fill[1]},{self._fill[2]})",
                    "stroke": f"rgb({self._stroke[0]},{self._stroke[1]},{self._stroke[2]})",
                    "stroke-width": self._stroke_weight,
                    "transform": f"rotate({self._rotation}, {x}, {y})",  # "translate(30,40) rotate(45)'
                }
                svg_element = SVG.tag("rect", attributes=attributes, self_close=True)
                self._svg_commands.append(svg_element)
        else:
            self.regular_polygon(x, y, width, height, 4, 45)

    def text(self, letters, x, y):
        if not isinstance(letters, str):
            letters = str(letters)

        param_types = {
            "x": [[int, float], []],
            "y": [[int, float], []],
            "letters": [str, []],
        }
        verify_func_param(self.text, param_types, locals())

        font = pygame.font.SysFont(self.sysfont, self._font_size)
        text = font.render(letters, True, self._fill)
        text = pygame.transform.rotate(text, self._rotation)
        self.screen.blit(text, (x, y))

        if self._svg:
            attributes = {
                "x": x,
                "y": y,
                "font-size": self._font_size,
                # 'fill': f'rgb({self._fill[0]},{self._fill[1]},{self._fill[2]})',
                # 'stroke': f'rgb({self._stroke[0]},{self._stroke[1]},{self._stroke[2]})',
                # 'stroke-width': self.stroke_size,
                "transform": f"rotate({self._rotation}, {x}, {y})",  # "translate(30,40) rotate(45)'
            }
            svg_element = SVG.tag("text", content=letters, attributes=attributes)
            self._svg_commands.append(svg_element)

    def arc(self, x, y, width, height, start_angle, end_angle):
        param_types = {
            "x": [[int, float], []],
            "y": [[int, float], []],
            "width": [[int, float], []],
            "height": [[int, float], []],
            "start_angle": [int, []],
            "end_angle": [int, []],
        }
        verify_func_param(self.arc, param_types, locals())
        pygame.draw.arc(
            self.screen,
            self._fill,
            [x, y, width, height],
            start_angle,
            end_angle,
            self._stroke_weight,
        )

    def begin_shape(self):
        self._polygon_coords = []

    def vertex(self, coord):
        self._polygon_coords.append(coord)

    def end_shape(self, fill=1):
        if fill:
            pygame.draw.polygon(self.screen, self._fill, self._polygon_coords)
            svg_fill = f"rgb({self._fill[0]},{self._fill[1]},{self._fill[2]})"
        else:
            pygame.draw.polygon(
                self.screen, self._fill, self._polygon_coords, self._stroke_weight
            )
            svg_fill = "none"

        if self._svg:
            points = []
            for p in self._polygon_coords:
                points.append(f"{p[0]},{p[1]}")
            attributes = {
                "points": " ".join(points),
                "fill": svg_fill,
                "stroke": f"rgb({self._stroke[0]},{self._stroke[1]},{self._stroke[2]})",
                "stroke-width": self._stroke_weight,
                "transform": f"rotate({self._rotation})",  # "translate(30,40) rotate(45)'
            }
            svg_element = SVG.tag("polygon", attributes=attributes, self_close=True)
            self._svg_commands.append(svg_element)

    def polygon(self, coords, fill=1):
        if fill:
            pygame.draw.polygon(self.screen, self._fill, coords)
            svg_fill = f"rgb({self._fill[0]},{self._fill[1]},{self._fill[2]})"
        else:
            pygame.draw.polygon(self.screen, self._fill, coords, self._stroke_weight)
            svg_fill = "none"

        if self._svg:
            points = []
            for p in self._polygon_coords:
                points.append(f"{p[0]},{p[1]}")
            attributes = {
                "points": " ".join(points),
                "fill": svg_fill,
                "stroke": f"rgb({self._stroke[0]},{self._stroke[1]},{self._stroke[2]})",
                "stroke-width": self._stroke_weight,
                "transform": f"rotate({self._rotation})",  # "translate(30,40) rotate(45)'
            }
            svg_element = SVG.tag("polygon", attributes=attributes, self_close=True)
            self._svg_commands.append(svg_element)

    def line(self, x1, y1, x2, y2):
        param_types = {
            "x1": [[int, float], []],
            "y1": [[int, float], []],
            "x2": [[int, float], []],
            "y2": [[int, float], []],
        }
        verify_func_param(self.line, param_types, locals())
        pygame.draw.line(
            self.screen, self._stroke, [x1, y1], [x2, y2], self._stroke_weight
        )

        if self._svg:
            attributes = {
                "x1": x1,
                "x2": x2,
                "y1": y1,
                "y2": y2,
                "stroke": f"rgb({self._stroke[0]},{self._stroke[1]},{self._stroke[2]})",
                "stroke-width": self._stroke_weight,
                "transform": f"rotate({self._rotation}, {x1}, {y1})",  # "translate(30,40) rotate(45)'
            }
            svg_element = SVG.tag("line", attributes=attributes, self_close=True)
            self._svg_commands.append(svg_element)

    def star(self, x, y, r1, r2, npoints):
        self._star(self, x, y, r1, r2, npoints, self._rotation)

    def alpha_ellipse(self, x, y, w, h):
        param_types = {
            "x": [[int, float], []],
            "y": [[int, float], []],
            "w": [int, []],
            "h": [int, []],
        }
        verify_func_param(self.alpha_ellipse, param_types, locals())
        self._alpha_ellipse(self, x, y, w, h)

    def curve_rect(self, x, y, w, h, curve):
        self._curve_rect(self, x, y, w, h, curve, self._rotation)

    def arrow(self, x, y, width, height):
        self._arrow(self, x, y, width, height, self._rotation)

    def heart(self, x, y, w, h):
        self._heart(self, x, y, w, h, self._rotation)

    def regular_polygon(self, x, y, w, h, num_of_points, angle_offset=0):
        self._reg_poly(self, x, y, w, h, num_of_points, self._rotation, angle_offset)

    def supershape(self, x_coord, y_coord, size_x, size_y, param_options, fill=False):
        self._supershape(
            self,
            x_coord,
            y_coord,
            size_x,
            size_y,
            param_options,
            self._rotation,
            fill=False,
        )

    def smooth_star(self, x_coord, y_coord, size_x, size_y, n1=0.20, fill=False):
        self._smooth_star(self, x_coord, y_coord, size_x, size_y, n1=n1, fill=fill)

    def oil_drop(self, x_coord, y_coord, size_x, size_y, n1=0.3, fill=False):
        self._oil_drop(self, x_coord, y_coord, size_x, size_y, n1, fill=fill)

    def flowing_star(self, x_coord, y_coord, size_x, size_y, n1=0.3, fill=False):
        self._flowing_star(self, x_coord, y_coord, size_x, size_y, n1, fill=fill)

    def manual_ellipse(self, x, y, w, h, a):
        ellipse(self, x, y, w, h, self._rotation, a)

    def gradient_rect(self, x, y, w, h, start_col, end_col, direction=0):
        val = w if direction == 0 else h
        val = 1 if val == 0 else val
        sr, sg, sb = start_col
        er, eg, eb = end_col
        dr, dg, db = (er - sr) / val, (eg - sg) / val, (eb - sb) / val
        if direction == 0:
            surf = pygame.Surface((w, 1))
        else:
            surf = pygame.Surface((1, h))
        for i in range(val):
            col = (int(sr + dr * i), int(sg + dg * i), int(sb + db * i))
            if direction == 0:
                surf.set_at((i, 0), col)
            else:
                surf.set_at((0, i), col)
        self.screen.blit(pygame.transform.scale(surf, (w, h)), (x, y))

    def fill_arc(self, x, y, radius, startangle, endangle, start_rad=0):
        for r in range(start_rad, radius):
            pygame.gfxdraw.arc(
                self.screen, x, y, r, int(startangle), int(endangle), self._fill
            )

    #
    # interactivity
    #

    def mouseX(self):
        x, y = pygame.mouse.get_pos()
        return x

    def mouseY(self):
        x, y = pygame.mouse.get_pos()
        return y

    def cross_hair(self, coord):
        self._cross_hair(self, coord)

    #
    # pygame
    #

    def flip_display(self):
        """updates the screen. This should be called once every frame"""
        pygame.display.flip()
        if self.bg_col is not None:
            self.background(self.bg_col)

    def handle_events(self, event):
        if event.type == pygame.QUIT:
            self.is_running = False

    def event_loop(self):
        """Get all new events. This should be called once every frame"""
        if len(self._timers) > 0:
            self._timer_update()
        self.mouse_test_x = self.mouseX()
        for event in pygame.event.get():
            self.handle_events(event)

    #
    # ui
    #

    def button(self, *args, **kwargs) -> Button:
        b = Button(*args, **kwargs)
        self._all_widgets.append(b)
        return b

    def text_box(self, *args, **kwargs) -> TextBox:
        t = TextBox(*args, **kwargs)
        self._all_widgets.append(t)
        return t

    def slider(self, *args, **kwargs) -> Slider:
        s = Slider(self, *args, **kwargs)
        self._all_widgets.append(s)
        return s

    def update_ui(self):
        for widget in self._all_widgets:
            widget.update()

    def slider_with_text(self, slider, params={}) -> slider_with_text:
        s = slider_with_text(self, slider, params)
        self._all_widgets.append(s)
        return s

    def scroll(self, param_options={}) -> Scroll:
        s = Scroll(self, param_options)
        self._all_widgets.append(s)
        return s

    #
    # time
    #

    def timer(self, callback=None, seconds=0, minutes=0) -> Timer:
        t = Timer(callback, seconds, minutes)
        self._timers.append(t)
        return t

    def _timer_update(self):
        l = len(self._timers) - 1
        for i, timer in enumerate(reversed(self._timers)):
            t = timer.update()
            if t:
                del self._timers[l - i]

    def hour(self) -> int:
        now = datetime.datetime.now()
        return now.hour

    def minute(self) -> int:
        now = datetime.datetime.now()
        return now.minute

    def second(self) -> int:
        now = datetime.datetime.now()
        return now.second

    #
    # charts
    #

    def barchart(self, x, y, w, h, params={}, **kwargs):
        params.update(kwargs)
        self._barchart(self, x, y, w, h, params)

    def linechart(self, x, y, w, h, params={}, **kwargs):
        params.update(kwargs)
        self._linechart(self, x, y, w, h, params)

    def piechart(self, x, y, radius, data, start_rad=0):
        """
        data in the format:
            [
            ['a', 20, hapi.color['red']],
            ['b', 30, hapi.color['blue']],
            ['c', 40, hapi.color['yellow']],
            ['d', 60, hapi.color['green']],
            ['e', 30, hapi.color['black']]
        ]
        """
        self._piechart(self, x, y, radius, data, start_rad=start_rad)

    def scatterchart(self, x, y, width, height, params={}, **kwargs):
        params.update(kwargs)
        self._scatterchart(self, x, y, width, height, params)

    #
    # svg
    #

    def save_svg(self, path):
        # print(self._svg_commands)
        SVG.save(self._svg_commands, path, self.WIDTH, self.HEIGHT)


    #
    # img
    #


    def save(self, path):
        self.pygame.image.save(self.screen, path)


    #
    # video
    #

    def record(self):
        if 'record_counter' not in self._session_vars:
            self._session_vars['record_counter'] = 1

        if self._session_vars['record_counter'] == 1:
            try:
                shutil.rmtree( 'hoomanvid' )
            except FileNotFoundError:
                pass

        try:
            os.mkdir('hoomanvid')
        except:
            pass
        self.save(f"hoomanvid/{self._session_vars['record_counter']}.png")
        self._session_vars['record_counter'] += 1

    def save_record(self, path):
        
        try:
            import ffmpeg
        except:
            try:
                shutil.rmtree( 'hoomanvid' )
            except FileNotFoundError:
                pass
            sys.exit("You need to pip install ffmpeg-python and ensure ffmpeg is installed to use this feature")
        
        (
            ffmpeg
            .input('hoomanvid/*.png', pattern_type='glob', framerate=25)
            .output(path)
            .run()
        )

        try:
            shutil.rmtree( 'hoomanvid' )
        except FileNotFoundError:
            pass

        



"""
Author: https://github.com/TheBigKahuna353
Edit: https://github.com/Abdur-rahmaanJ
"""

import pygame
from .formula import constrain
from typing import Union
from .check import check_params


class Button:
    def __init__(self, x, y, w, h, text, params={}):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.text = text

        options = {
            "surface": None,
            "hover_background_color": None,
            "font": "Calibri",
            "font_size": 30,
            "outline": False,
            "outline_thickness": 2,
            "outline_color": (0, 0, 0),
            "outline_half": False,
            "on_click": None,
            "on_hover_enter": None,
            "on_hover_exit": None,
            "on_hover": None,
            "on_hold": None,
            "image": None,
            "hover_image": None,
            "enlarge": False,
            "enlarge_amount": 1.1,
            "calculate_size": False,
            "dont_generate": False,
            "font_colour": (0, 0, 0),
            "background_color": (255, 255, 255),
            "curve": 0,
            "padding_x": 0,
            "padding_y": 0,
            "centered": False,
        }
        check_params(params, options, "button")
        options.update(params)

        for key, val in options.items():
            if key not in options:
                raise TypeError(key + " is not an option, have you spelt it correctly")

        self.padding_x = options["padding_x"]
        self.padding_y = options["padding_y"]
        self.surface = options["surface"]
        self.text_colour = options["font_colour"]
        self.background_color = options["background_color"]
        self.hover_bg_colour = options["hover_background_color"]
        self.curve = options["curve"]
        font = options["font"]
        font_size = options["font_size"]
        self.outline = options["outline"]
        self.outline_col = options["outline_color"]
        self.outline_half = options["outline_half"]
        self.outline_amount = options["outline_thickness"]
        self.on_click = options["on_click"]
        self.on_hover_enter = options["on_hover_enter"]
        self.on_hover_exit = options["on_hover_exit"]
        self.on_hover = options["on_hover"]
        self.on_hold = options["on_hold"]
        image = options["image"]
        dont_generate = options["dont_generate"]
        self.caclulateSize = options["calculate_size"]
        self.hover_image = options["hover_image"]
        self.enlarge = options["enlarge"]
        self.enlarge_amount = options["enlarge_amount"]
        self.center = options["centered"]

        # if no surface is supplied, try getting main screen
        if self.surface is None:
            self.surface = pygame.display.get_surface()
            if self.surface is None:
                raise ValueError("No surface to blit to")

        if self.hover_bg_colour is None:
            self.hover_bg_colour = self.background_color
        
        self.font = pygame.font.Font(pygame.font.match_font(font), font_size)

        self.image = image.copy() if image else None
        self.clicked_on = False
        self.draw = False

        if self.enlarge:
            if self.text != "":
                self.enlarge_font = pygame.font.Font(
                    pygame.font.match_font(font), int(font_size * self.enlarge_amount)
                )
        self.hover = False

        self.prev_clicked_state = False
        # create the surfaces for the button to blit every frame
        if not dont_generate:
            if self.w == 0 or self.h == 0 or self.caclulateSize:
                if image is not None:
                    self.w = self.image.get_width()
                    self.h = self.image.get_height()
                else:
                    if self.text != "":
                        self._caclulate_size()
                    else:
                        raise ValueError(
                            "cannot calculate width and height without text"
                        )
            self._Generate_images()
        if self.center:
            self.x -= self.w // 2
            self.y -= self.h // 2

    def _Generate_images(self):
        # generate images
        # if no image, create the button by drawing
        if self.image is None or self.draw:
            self.draw = True
            self.image = pygame.Surface((self.w, self.h), pygame.SRCALPHA)
            self.hover_image = pygame.Surface((self.w, self.h), pygame.SRCALPHA)
            self.image.blit(
                curve_square(self.w, self.h, self.curve, self.background_color), (0, 0)
            )
            self.hover_image.blit(
                curve_square(self.w, self.h, self.curve, self.hover_bg_colour), (0, 0)
            )
            # self.hover_image.fill(self.background_color)
            if self.outline:
                self.hover_image.blit(
                    curve_square(self.w, self.h, self.curve, self.outline_col), (0, 0)
                )
                self.hover_image.blit(
                    curve_square(
                        self.w - self.outline_amount * 2,
                        self.h - self.outline_amount * 2,
                        self.curve,
                        self.hover_bg_colour,
                    ),
                    (self.outline_amount, self.outline_amount),
                )
            elif self.outline_half:
                self.hover_image.blit(
                    curve_square(self.w, self.h, self.curve, self.outline_col), (0, 0)
                )
            self.hover_image.convert()
            self.image.convert()
        # if the user gives an image, create the image when the mouse hovers over
        elif self.hover_image is None or self.draw:
            self.hover_image = self.image.copy()
            if not self.outline is None:
                pygame.draw.rect(
                    self.hover_image,
                    (0, 0, 0, 255),
                    (0, 0, self.w, self.outline_amount),
                )
                pygame.draw.rect(
                    self.hover_image,
                    (0, 0, 0, 255),
                    (0, 0, self.outline_amount, self.h),
                )
                pygame.draw.rect(
                    self.hover_image,
                    (0, 0, 0, 255),
                    (self.w, self.h, -self.w, -self.outline_amount),
                )
                pygame.draw.rect(
                    self.hover_image,
                    (0, 0, 0, 255),
                    (self.w, self.h, -self.outline_amount, -self.h),
                )
            self.hover_image.convert_alpha()
            self.image.convert_alpha()
        # enlarge the image, no matter if user gives an image or not
        if self.enlarge:
            size = (
                int(self.w * self.enlarge_amount),
                int(self.h * self.enlarge_amount),
            )
            self.dx, self.dy = size[0] - self.w, size[1] - self.h
            self.hover_image = pygame.transform.scale(self.image, size)
        # put the text over images, if enlarge, create a bigger font so resolution stays high
        if self.text != "":
            txt = self.font.render(self.text, True, self.text_colour)
            self.image.blit(
                txt, ((self.w - txt.get_width()) // 2, (self.h - txt.get_height()) // 2)
            )
            if self.enlarge:
                txt = self.enlarge_font.render(self.text, True, self.text_colour)
            self.hover_image.blit(
                txt,
                (
                    (self.hover_image.get_width() - txt.get_width()) // 2,
                    (self.hover_image.get_height() - txt.get_height()) // 2,
                ),
            )
        # if the user gives both images, check to see if different sizes so know if enlarged or not
        if (
            self.hover_image.get_width() != self.w
            or self.hover_image.get_height() != self.h
        ):
            self.enlarge = True
            self.dx, self.dy = (
                self.hover_image.get_width() - self.w,
                self.hover_image.get_height() - self.h,
            )
        # convert the images so it is faster to put on screen
        self.image.convert()
        self.hover_image.convert()

    # if no width or height is given, calculate it with length of text
    def _caclulate_size(self):
        txt = self.font.render(self.text, False, (0, 0, 0))
        self.w = txt.get_width() + self.w + self.padding_x * 2
        self.h = txt.get_height() + self.h + self.padding_y * 2

    # return a pygame.Rect of the button
    def get_rect(self) -> pygame.Rect:
        return pygame.Rect(self.x, self.y, self.w, self.h)

    # this is what will be shown when print(button)
    def __str__(self):
        if self.text:
            return "Button: '" + self.text + "'"
        else:
            return "Button: at (" + str(self.x) + ", " + str(self.y) + ")"

    # update the text of the button, remake the surfaces for the button
    def Update_text(self, text):
        self.text = text
        if self.caclulateSize:
            self._caclulate_size()
        self._Generate_images()

    # update the button, this should get called every frame
    def update(self) -> bool:
        click = pygame.mouse.get_pressed()[0]
        mouse_pos = pygame.mouse.get_pos()
        returnee = False
        # check if mouse over button
        if mouse_pos[0] > self.x and mouse_pos[0] < self.x + self.w:
            if mouse_pos[1] > self.y and mouse_pos[1] < self.y + self.h:
                # on_hover_enter event
                if self.hover == False:
                    if self.on_hover_enter:
                        self.on_hover_enter(self)

                self.hover = True
                # on_hover event
                if self.on_hover:
                    self.on_hover(self)

                # check for click, if held down, action only gets called once AT MOUSEUP

                #check to see if the click started on the button
                if click and not self.prev_clicked_state:
                    self.clicked_on = True

                # wait till the user stops clicking to call on_click
                if self.prev_clicked_state and self.clicked_on and click == False:
                    if self.on_click:
                        """
                        if self.on_click_arg:
                            self.on_click(self.on_click_arg)
                        else:
                            self.on_click()
                        """
                        self.on_click(self)
                    returnee = True
                if not click:
                    self.clicked_on = False
            else:
                if self.hover:
                    if self.on_hover_exit:
                        self.on_hover_exit(self)
                self.hover = False
        else:
            if self.hover:
                if self.on_hover_exit:
                    self.on_hover_exit(self)
            self.hover = False
        self.prev_clicked_state = click
        # draw
        self._draw()
        # return if the button was clicked on
        return returnee

    def create_button(self):
        self._Generate_images()

    # draw the button
    def _draw(self):
        if self.hover:
            if self.enlarge:
                self.surface.blit(
                    self.hover_image, (self.x - self.dx // 2, self.y - self.dy // 2)
                )
            else:
                self.surface.blit(self.hover_image, (self.x, self.y))
        else:
            self.surface.blit(self.image, (self.x, self.y))

    def width(self) -> int:
        txt = self.font.render(self.text, False, (0, 0, 0))
        w = txt.get_width() + self.w + self.padding_x * 2
        return w

    def height(self) -> int:
        txt = self.font.render(self.text, False, (0, 0, 0))
        h = txt.get_height() + self.h + self.padding_y * 2
        return


# this creates a curved rect, given a w,h and the curve amount, bewtween 0 and 1
def curve_square(width, height, curve, color=(0, 0, 0)):
    if not 0 <= curve <= 1:
        raise ValueError("curve value out of range, must be between 0 and 1")
    curve /= 2
    curve *= min(width, height)
    curve = int(curve)
    surf = pygame.Surface((width, height), pygame.SRCALPHA)
    pygame.draw.rect(surf, color, (0, curve, width, height - 2 * curve))
    pygame.draw.rect(surf, color, (curve, 0, width - 2 * curve, height))
    pygame.draw.circle(surf, color, (curve, curve), curve)
    pygame.draw.circle(surf, color, (width - curve, curve), curve)
    pygame.draw.circle(surf, color, (curve, height - curve), curve)
    pygame.draw.circle(surf, color, (width - curve, height - curve), curve)
    return surf


class Slider:
    def __init__(self, hapi, x, y, w, h, params={}):
        options = {
            "background_color": (100, 100, 100),
            "slider_width": None,
            "slider_color": (200, 200, 200),
            "starting_value": None,
            "value_range": [0, 1],
            "slider_height": None,
            "step": 0,
            "image": None,
            "direction": "horizontal",
            "resize_slider": False,
            "curve": 0,
        }
        check_params(params, options, "slider")
        options.update(params)

        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.hapi = hapi
        self.direction = options["direction"]
        if self.direction not in ["horizontal", "vertical"]:
            raise ValueError(
                "option 'direction' is not a direction, (%d)" % (self.direction)
            )
        self.bg = options["background_color"]
        # if the user gives an image for slider background, use that instead of drawing one
        if options["image"] is not None:
            self.image = pygame.Surface((self.w, self.h))
            self.image.blit(options["image"], (0, 0))
        else:
            self.image = None
        self.val_range = options["value_range"]
        self.curve = options["curve"]
        self.resize = options["resize_slider"]
        val_dif = self.val_range[1] - self.val_range[0]
        self.slider_bg = options["slider_color"]
        self.slider_h = options["slider_height"]
        if self.slider_h is None:
            if self.direction == "horizontal":
                self.slider_h = h
            else:
                self.slider_h = w
        self.step = options["step"]
        if self.direction == "horizontal":
            self.slider_w = (
                options["slider_width"] if options["slider_width"] is not None else h
            )
        else:
            self.slider_w = (
                options["slider_width"] if options["slider_width"] is not None else w
            )
        if options["starting_value"] is not None:
            self.val = constrain(
                options["starting_value"], self.val_range[0], self.val_range[1], 0, 1
            )
        else:
            self.val = 0.5
        if self.resize:
            range_ = self.val_range[1] - self.val_range[0]
            if self.direction == "horizontal":
                if range_ < self.w:
                    self.slider_w = self.w - range_
            else:
                if range_ < self.h:
                    self.slider_h = self.h - range_
        self.screen = pygame.display.get_surface()
        if self.direction == "horizontal":
            self.slider_rect = pygame.Rect(
                self.x + self.val * (self.w - self.slider_w),
                self.y + (self.h - self.slider_h) // 2,
                self.slider_w,
                self.slider_h,
            )
        else:
            self.slider_rect = pygame.Rect(
                self.x + (self.w - self.slider_w) // 2,
                self.y + self.val * (self.h - self.slider_h),
                self.slider_w,
                self.slider_h,
            )
        self.clicked_on = False
        self.prev_click = False

    # draw the slider
    def _draw(self):
        if self.image is not None:
            self.hapi.screen.blit(self.image, (self.x, self.y))
        elif self.curve == 0:
            pygame.draw.rect(
                self.hapi.screen, self.bg, (self.x, self.y, self.w, self.h)
            )
        else:
            self.hapi.screen.blit(
                curve_square(self.w, self.h, self.curve, self.bg), (self.x, self.y)
            )
        if self.curve == 0:
            pygame.draw.rect(self.hapi.screen, self.slider_bg, self.slider_rect)
        else:
            self.hapi.screen.blit(
                curve_square(self.slider_w, self.slider_h, self.curve, self.slider_bg),
                self.slider_rect,
            )

    # updates the slider, this should be called every frame
    def update(self):
        mouse_pos = pygame.mouse.get_pos()
        click = pygame.mouse.get_pressed()[0]
        if self.slider_rect.collidepoint(mouse_pos):
            # check for click, if held down, action only gets called once
            if click and not self.prev_click:
                self.clicked_on = True
        if self.clicked_on:
            if self.direction == "horizontal":
                self.val = (mouse_pos[0] - self.x) / self.w
                self.val = max(min(self.val, 1), 0)
                self.val = self._get_val(self.val)
                self.slider_rect.x = self.x + self.val * (self.w - self.slider_w)
            else:
                self.val = (mouse_pos[1] - self.y) / self.h
                self.val = max(min(self.val, 1), 0)
                self.val = self._get_val(self.val)
                self.slider_rect.y = self.y + self.val * (self.h - self.slider_h)
        if not click:
            self.clicked_on = False
        self.prev_click = click
        self._draw()

    def Move(self, x=0, y=0, dx=0, dy=0):
        self.x = x if x != 0 else self.x
        self.y = y if y != 0 else self.y
        self.x += dx
        self.y += dy
        if self.direction == "horizontal":
            self.slider_rect = pygame.Rect(
                self.x + self.val * (self.w - self.slider_w),
                self.y + (self.h - self.slider_h) // 2,
                self.slider_w,
                self.slider_h,
            )
        else:
            self.slider_rect = pygame.Rect(
                self.x + (self.w - self.slider_w) // 2,
                self.y + self.val * (self.h - self.slider_h),
                self.slider_w,
                self.slider_h,
            )

    #returns the value the slider is at
    def value(self) -> int:
        val = constrain(self.val, 0, 1, self.val_range[0], self.val_range[1])
        if isinstance(self.step, int) and self.step != 0:
            val = int(val)
        return val

    #sets the value of the slider, moveing the slider object to that position
    def set_value(self, val: Union[float, int]):
        self.val = constrain(val, self.val_range[0], self.val_range[1], 0, 1)
        if self.direction == "horizontal":
            self.slider_rect.x = self.x + self.val * (self.w - self.slider_w)
        else:
            self.slider_rect.y = self.y + self.val * (self.h - self.slider_h)

    # if the slider has a step and is not contineous, round to nearest step size
    def _get_val(self, val):
        if self.step == 0:
            return val
        else:
            a = constrain(val, 0, 1, self.val_range[0], self.val_range[1])
            b = self.hapi.round_to(a, self.step)
            c = constrain(b, self.val_range[0], self.val_range[1], 0, 1)
            # print(val, a, b, c)
            return c


class TextBox:
    def __init__(self, x, y, w, h=0, params={}):
        options = {
            "max_lines": 1000,
            "text": "",
            "background_color": (255, 255, 255),
            "font_size": 30,
            "font": "Calibri",
            "text_colour": (0, 0, 0),
            "surface": None,
            "margin": 2,
            "cursor": True,
            "on_enter": None,
            "calculate_size": False,
        }
        check_params(params, options, "text box")
        options.update(params)
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        self.cursor = options["cursor"]
        self.current_line = 0
        self.current_col = len(options["text"])
        self.lines = options["lines"]
        self.font = pygame.font.Font(
            pygame.font.match_font(options["font"]), options["font_size"]
        )
        self.text_colour = options["text_colour"]
        self.text = [list(options["text"])]
        self.char_length = [self._get_text_width(x) for x in self.text]
        self.background = options["background_color"]
        self.surface = (
            options["surface"]
            if options["surface"] is not None
            else pygame.display.get_surface()
        )
        self.margin = options["margin"]
        self.Enter_action = options["on_enter"]
        if self.surface == None:
            raise ValueError("No surface to blit to")
        # if no surface is supplied, get window
        if self.surface == None:
            self.surface = pygame.display.get_surface()
            if self.surface == None:
                raise ValueError("No surface to blit to")
        if options["calculateSize"] or self.h == 0:
            self.h = self._get_font_height() + h

    # get the width of the text using the font
    def _get_text_width(self, text):
        text = "".join(text)
        if len(text) == 0:
            return 0
        obj = self.font.render(text, True, (0, 0, 0))
        return obj.get_width()

    # returns the height of the font
    def _get_font_height(self):
        obj = self.font.render(" ", True, (0, 0, 0))
        return obj.get_height()

    def wrapper(self, change_cur=False):
        for cur_line, line in enumerate(self.text):
            for i in range(len("".join(line))):
                length = self._get_text_width("".join(line[:i]))
                if length > self.w:
                    indexs = [
                        i for i, e in enumerate(self.text[cur_line][:i]) if e == " "
                    ]
                    if cur_line < self.lines - 1:
                        if len(indexs) == 0:
                            indexs.append(i - 1)
                        if change_cur:
                            self.current_line += 1
                            self.current_col = len(self.text[cur_line]) - indexs[-1] - 1
                        if cur_line < len(self.text):
                            self.text.append(self.text[cur_line][indexs[-1] + 1 :])
                        else:
                            self.text[cur_line + 1] = (
                                self.text[cur_line][indexs[-1] + 1 :]
                                + self.text[cur_line]
                            )
                        self.text[cur_line] = self.text[cur_line][: indexs[-1]]
                        break

    # call this when the user presses a key down, supply the event from `pygame.event.get()`
    def key_down(self, e: pygame.event.Event):
        # when backspace is pressed, delete last char
        if e.key == pygame.K_BACKSPACE:
            # if nothing in line, delete line
            if len(self.text[self.current_line]) == 0:
                if self.current_line > 0:
                    del self.text[self.current_line]
                    self.current_line -= 1
                    self.current_col = len(self.text[self.current_line])
            else:
                del self.text[self.current_line][-1]
                self.current_col -= 1
        # if key is enter, create line
        elif e.key == pygame.K_RETURN:
            if self.Enter_action is not None:
                self.Enter_action()
            elif self.current_line < self.lines - 1:
                self.current_line += 1
                self.text.append([""])
                self.char_length.append([0])
                self.current_col = 0
        # if key is a charachter, put on screen
        elif e.unicode != "":
            if len(self.text[self.current_line]) > 0:
                if self.text[self.current_line][-1] == "":
                    del self.text[self.current_line][-1]
            self.text[self.current_line] = (
                self.text[self.current_line][: self.current_col]
                + [e.unicode]
                + self.text[self.current_line][self.current_col :]
            )
            self.current_col += 1
            # wrapper
            if self._get_text_width(self.text[self.current_line]) > self.w:
                self.wrapper(True)
        # if the down arrow is pressed
        elif e.key == pygame.K_DOWN:
            self.current_line += 1 if self.current_line < len(self.text) - 1 else 0
            self.current_col = min(self.current_col, len(self.text[self.current_line]))
        # if the up arrow is pressed
        elif e.key == pygame.K_UP:
            self.current_line -= 1 if self.current_line > 0 else 0
            self.current_col = min(self.current_col, len(self.text[self.current_line]))
        # if the right arrow is pressed
        elif e.key == pygame.K_RIGHT:
            self.current_col += (
                1 if len(self.text[self.current_line]) > self.current_col else 0
            )
        # if the left arrow is pressed
        elif e.key == pygame.K_LEFT:
            self.current_col -= 1 if 0 < self.current_col else 0

    # draw the textbox
    def _draw(self):
        # draw background
        pygame.draw.rect(
            self.surface, self.background, (self.x, self.y, self.w, self.h * self.lines)
        )
        # draw all text
        for line, text in enumerate(self.text):
            if len(text) != 0:
                txt = "".join(text)
                obj = self.font.render(txt, True, self.text_colour)
                self.surface.blit(obj, (self.x + self.margin, self.y + (self.h * line)))
        # draw cursor
        if self.cursor:
            total = 0
            total = self._get_text_width(
                self.text[self.current_line][: self.current_col]
            )
            pygame.draw.line(
                self.surface,
                (0, 0, 0),
                (self.x + total, self.y + (self.h * self.current_line)),
                (self.x + total, self.y + self._get_font_height() + (self.h * (self.current_line))),
                2,
            )

    # update should be called every frame, it draws the textbox
    def update(self):
        self._draw()

    # get the text of a specific line or lines
    def get_lines(self, lines=-1, return_as_string=False):
        pas = False
        # if user gives an int, check if it is -1 for all lines, else get specific line
        if isinstance(lines, int):
            if lines == -1:
                lines = (0, self.lines)
                pas = True
            if not pas:
                if 0 > lines or self.lines < lines:
                    raise IndexError("line index not in range")
                if len(self.text) < lines:
                    return ""
                return "".join(self.text[lines])
        # if user wants a range of lines, get lines
        if isinstance(lines, tuple):
            if (
                lines[0] < 0
                or lines[0] > self.lines
                or lines[1] < 0
                or lines[1] > self.lines
                or lines[0] > lines[1]
            ):
                raise IndexError(
                    "line index is out of range: "
                    + str(lines)
                    + " (0, "
                    + str(str(self.lines))
                )
            string = []
            for x in range(lines[0], lines[1]):
                if len(self.text) > x:
                    string.append("".join(self.text[x]))
                else:
                    string.append("")
            if return_as_string:
                return "\n".join(string)
            return string


class slider_with_text:
    def __init__(self, hapi, slider, params={}):
        options = {
            "font": "calibri",
            "font_size": 20,
            "font_color": (0, 0, 0),
            "padding_y": 2,
            "padding_x": 0,
            "pivot": "top_left",
            "accuracy": 0,  # decimal points
        }

        if not isinstance(slider, Slider):
            raise TypeError("'Slider' argument is not a slider widget")

        check_params(params, options, "slider with text")

        options.update(params)

        self.font = pygame.font.Font(
            pygame.font.match_font(options["font"]), options["font_size"]
        )
        self.font_col = options["font_color"]
        self.rounder = options["accuracy"]
        self.pivot = options["pivot"]
        if self.pivot not in [
            "top_left",
            "top_right, bottom_left",
            "bottom_right",
            "top",
            "bottom",
            "left",
            "right",
        ]:
            raise TypeError("option: 'pivot' is not a valid pivot point")

        self.slider = slider
        self.hapi = hapi
        self.y = slider.y
        self.val = 0
        if "top" in self.pivot:
            self.y -= options["padding_y"] + self._get_text_height()
        elif "bottom" in self.pivot:
            self.y += options["padding_y"] + slider.h
        self.x = slider.x
        if "left" in self.pivot:
            self.x -= options["padding_x"]
        elif "right" in self.pivot:
            self.x += slider.w + options["padding_x"]
        else:
            self.x += slider.w // 2 + options["padding_x"]

    def update(self):
        self.slider.update()
        self.val = round(self.slider.value(), self.rounder)
        if self.rounder == 0:
            self.val = int(self.val)
        self._draw()

    def _draw(self):
        val = str(self.val)
        obj = self.font.render(val, True, self.font_col)
        if "left" in self.pivot:
            x = self.x - obj.get_width()
        elif not "right" in self.pivot:
            x = self.x - obj.get_width() // 2
        else:
            x = self.x
        if "top" in self.pivot:
            y = self.y - obj.get_height()
        y = self.y
        self.hapi.screen.blit(obj, (x, y))

    def _get_text_height(self):
        return self.font.render(" ", True, (0, 0, 0)).get_height()

    def value(self) -> int:
        return self.slider.value()


class Scroll:
    def __init__(self, hapi, params={}):
        options = {
            "starting_x": 0,
            "starting_y": 0,
            "range_x": 0,
            "range_y": 0,
            "bar_color": (200, 200, 200),
            "slider_color": (150, 150, 150),
        }
        check_params(params, options, "scroll")
        options.update(params)

        screen_size = pygame.display.get_surface().get_size()

        self.w, self.h = screen_size
        # create sliders for x and y axis
        self.x_slider = None
        self.y_slider = None
        if options["range_x"] > 0:
            self.x_slider = Slider(
                hapi,
                0,
                self.h - 20,
                self.w - 20,
                20,
                {
                    "starting_value": options["starting_x"],
                    "value_range": [0, options["range_x"]],
                    "slider_color": options["slider_color"],
                    "background_color": options["bar_color"],
                    "resize_slider": True,
                },
            )
        if options["range_y"] > 0:
            self.y_slider = Slider(
                hapi,
                self.w - 20,
                0,
                20,
                self.h - 20,
                {
                    "starting_value": options["starting_y"],
                    "value_range": [0, options["range_y"]],
                    "slider_color": options["slider_color"],
                    "background_color": options["bar_color"],
                    "direction": "vertical",
                    "resize_slider": True,
                },
            )

    # this is updates the sliders, this should be called every frame
    def update(self):
        if self.x_slider is not None:
            self.x_slider.update()
        if self.y_slider is not None:
            self.y_slider.update()
    
    #returns the scroll amount given an index e.g. 'scrollx = scroll[0]'
    def __getitem__(self, index : int):
        if index == 0:
            return -self.x_slider.value()
        elif index == 1:
            return -self.y_slider.value()
        return (-self.x_slider.value(), -self.y_slider.value())



import time


class Timer:
    def __init__(self, callback, seconds=0, minutes=0):
        self.callback = callback
        self.length = seconds + (minutes * 60)
        self.start_time = time.time()
        self.finsished = False

    def update(self):
        if time.time() - self.start_time >= self.length:
            self.finsished = True
            if self.callback is not None:
                self.callback()
            return True
        return False

    def __str__(self):
        return "Timer: %d seconds remaining" % (
            self.length - (time.time() - self.start_time)
        )

    def __bool__(self):
        return self.finsished



import pygame
import numpy

from math import pi
from math import cos
from math import sin
from math import radians
from math import sqrt

from pygame.gfxdraw import textured_polygon as poly


def star(hapi, x, y, r1, r2, npoints, rotation):
    """
    https://processing.org/examples/star.html
    """
    if npoints < 2:
        npoints = 2
    angle = hapi.PI * 2 / npoints
    half_angle = angle / 2
    hapi.begin_shape()
    rotation = radians(rotation)
    for a in numpy.arange(0 + rotation, hapi.PI * 2 + rotation, angle):
        sx = x + cos(a) * r2
        sy = y + sin(a) * r2
        hapi.vertex((sx, sy))
        sx = x + cos(a + half_angle) * r1
        sy = y + sin(a + half_angle) * r1
        hapi.vertex((sx, sy))
    hapi.end_shape()


def alpha_ellipse(hapi, x, y, w, h):
    # https://github.com/furas
    # https://stackoverflow.com/questions/59293057/how-to-make-transparent-pygame-draw-circle/
    surface1 = hapi.screen.convert_alpha()
    # surface1.fill([0,0,0,0])
    shape_fill = hapi._fill + (hapi._alpha,)
    pygame.draw.ellipse(surface1, shape_fill, (x, y, w, h))
    hapi.screen.blit(surface1, (0, 0))


def curve_rect(hapi, x, y, width, height, curve, rotation):
    curve /= 200
    curve = min(max(curve, 0), 1)
    curve *= min(width, height)
    curve = int(curve)

    shape_fill = hapi._fill + (hapi._alpha,)
    """
    surf = pygame.Surface((width, height), pygame.SRCALPHA)
    pygame.draw.rect(surf, shape_fill, (0, curve, width, height - 2 * curve))
    pygame.draw.rect(surf, shape_fill, (curve, 0, width - 2 * curve, height))
    pygame.draw.circle(surf, shape_fill, (curve, curve), curve)
    pygame.draw.circle(surf, shape_fill, (width - curve, curve), curve)
    pygame.draw.circle(surf, shape_fill, (curve, height - curve), curve)
    pygame.draw.circle(surf, shape_fill, (width - curve, height - curve), curve)
    hapi.screen.blit(surf,(x,y))
    """
    hapi.rect(x + curve, y, width - curve * 2, height)
    hapi.rect(x, y + curve, width, height - curve * 2)
    top_left = pygame.Vector2(-width // 2 + curve, -height // 2 + curve).rotate(
        rotation
    )
    top_right = pygame.Vector2(width // 2 - curve, -height // 2 + curve).rotate(
        rotation
    )
    bot_right = pygame.Vector2(width // 2 - curve, height // 2 - curve).rotate(rotation)
    bot_left = pygame.Vector2(-width // 2 + curve, height // 2 - curve).rotate(rotation)
    midx = x + width // 2
    midy = y + height // 2
    pygame.draw.circle(
        hapi.screen,
        shape_fill,
        (int(top_left[0]) + midx, int(top_left[1]) + midy),
        curve,
    )
    pygame.draw.circle(
        hapi.screen,
        shape_fill,
        (int(top_right[0]) + midx, int(top_right[1]) + midy),
        curve,
    )
    pygame.draw.circle(
        hapi.screen,
        shape_fill,
        (int(bot_left[0]) + midx, int(bot_left[1]) + midy),
        curve,
    )
    pygame.draw.circle(
        hapi.screen,
        shape_fill,
        (int(bot_right[0]) + midx, int(bot_right[1]) + midy),
        curve,
    )


def arrow(hapi, x, y, w, h, angle):
    """
    https://www.reddit.com/r/pygame/comments/glomfa/drawing_polygons/fr0ig8y/?context=3
    """
    start = pygame.Vector2(x + w, y + h)
    left = pygame.Vector2(-1 * w, -1 * h).rotate(angle)
    top = pygame.Vector2(0, -0.5 * h).rotate(angle)
    right = pygame.Vector2(1 * w, -1 * h).rotate(angle)
    down = pygame.Vector2(0, 0.5 * h).rotate(angle)
    hapi.begin_shape()
    hapi.vertex(start + left)
    hapi.vertex(start + top)
    hapi.vertex(start + right)
    hapi.vertex(start + down)
    hapi.end_shape()


def heart(hapi, x, y, w, h, rotation):
    """
    http://www.mathematische-basteleien.de/heart.htm
    """
    step_size = 1000 / (hapi.PI / 2)
    curve_h = max(int(h * 0.6), 1)
    circle_h = max(int(h * 0.4), 1)
    wr = max(w // 2, 1)
    hr = max(curve_h // 2, 1)
    start = pygame.Vector2(x + w, y + hr)
    hapi.begin_shape()
    for i in range(-1000, 1000, 1):
        if i == 0:
            continue
        d = pygame.Vector2(
            int(-wr + (sin(i / step_size) * wr)), circle_h // 2 + i // (1000 / hr)
        )
        hapi.vertex(start + d.rotate(rotation))
    hapi.vertex(start + pygame.Vector2(0, -hr + circle_h // 2).rotate(rotation))
    hapi.end_shape()

    hapi.begin_shape()
    for i in range(-1000, 1000, 1):
        if i == 0:
            continue
        d = pygame.Vector2(
            int(wr - (sin(i / step_size) * wr)), circle_h // 2 + i // (1000 / hr)
        )
        hapi.vertex(start + d.rotate(rotation))
    hapi.vertex(start + pygame.Vector2(0, -hr + circle_h // 2).rotate(rotation))
    hapi.end_shape()

    ellippse = start + pygame.Vector2(-w, -hr).rotate(rotation)
    ellippse1 = start + pygame.Vector2(0, -hr).rotate(rotation)
    hapi.ellipse(ellippse[0], ellippse[1], abs(w), abs(circle_h))
    hapi.ellipse(ellippse1[0], ellippse1[1], abs(w), abs(circle_h))
    # pygame.draw.circle(hapi.screen, (0,0,255), (int(start[0]), int(start[1])), 5)


def regular_polygon(hapi, x, y, w, h, n, rotation=0, angle_offset=0):
    if n < 3:
        n = 3

    midpoint = pygame.Vector2(x + w // 2, y + h // 2)
    # print(midpoint[0] - w//2, midpoint[1] - h//2)
    # r = sqrt((w/2)**2 + (h/2)**2)

    # if angle_offset != 0:
    # w = (w//2)//cos(angle_offset)
    # if angle_offset != 90:
    # h = (h//2)//sin(angle_offset)
    # w = r/2
    # h = 2*(r**2 - 0.5*w**2)
    # w = w/sin(angle_offset)
    # h = h/sin(angle_offset)
    # w = w/cos(angle_offset)
    # h = 2*h/cos(angle_offset)
    # w = (r/h) * w
    # h = (r/w) * h
    w -= 1
    h -= 1
    w *= sqrt(2)
    h *= sqrt(2)

    hapi.begin_shape()

    for angle in range(0, 360, 360 // n):
        angle = radians(angle + angle_offset)
        d = pygame.Vector2(-sin(angle) * w // 2, -cos(angle) * h // 2).rotate(rotation)

        hapi.vertex(midpoint + d)

    hapi.end_shape(hapi._stroke_weight)


def gradient_rect(w, h, start_col, end_col, direction=0):
    val = w if direction == 0 else h
    val = 1 if val == 0 else val
    sr, sg, sb = start_col
    er, eg, eb = end_col
    dr, dg, db = (er - sr) / val, (eg - sg) / val, (eb - sb) / val
    if direction == 0:
        surf = pygame.Surface((w, 1))
    else:
        surf = pygame.Surface((1, h))
    for i in range(val):
        col = (int(sr + dr * i), int(sg + dg * i), int(sb + db * i))
        if direction == 0:
            surf.set_at((i, 0), col)
        else:
            surf.set_at((0, i), col)
    return pygame.transform.scale(surf, (w, h))


def ellipse(hapi, x, y, w, h, rotation, a):

    midpoint = pygame.Vector2(x + w // 2, y + h // 2)

    resolution = 5

    points = []

    highest_h = 0
    highest_w = 0
    for angle in range(360 * resolution):
        angle = radians(angle / resolution)
        d = pygame.Vector2(sin(angle) * w // 2, cos(angle) * h // 2).rotate(rotation)
        highest_h = int(max(highest_h, d[1]))
        highest_w = int(max(highest_w, d[0]))

        points.append(midpoint + d)
    # print(highest_w, highest_h)
    poly(
        hapi.screen,
        points,
        gradient_rect(
            highest_w * 2, highest_h * 2, hapi.color["green"], hapi.color["red"]
        ),
        int(max(cos(-2 * hapi.PI * rotation / 180) * w - w, 0)),
        0,
    )


#
# Supershapes from http://paulbourke.net/geometry/supershape/
# With help from Daniel Shiefman
#


def r_val(theta, n1, n2, n3, m, a, b):
    if n1 == 0:
        n1 = 0.1
    part1 = (1 / a) * cos(theta * m / 4)
    part1 = abs(part1)
    part1 = pow(part1, n2)

    part2 = (1 / b) * sin(theta * m / 4)
    part2 = abs(part2)
    part2 = pow(part2, n3)

    part3 = pow(part1 + part2, 1 / n1)

    # returning r

    if part3 == 0:
        return 0

    return 1 / part3


def supershape(
    hapi, x_coord, y_coord, size_x, size_y, param_options, rotation, fill=False
):
    options = {"n1": 0.20, "n2": 1.7, "n3": 1.7, "m": 5, "a": 1, "b": 1, "phi": 2}

    pivot_x = x_coord
    pivot_y = y_coord

    options.update(param_options)

    n1 = options["n1"]
    n2 = options["n2"]
    n3 = options["n3"]
    m = options["m"]
    a = options["a"]
    b = options["b"]
    phi = options["phi"]

    rotation = radians(rotation)
    hapi.begin_shape()
    for angle in numpy.arange(0, hapi.PI * phi, 0.01):
        r = r_val(angle, n1, n2, n3, m, a, b)
        x = pivot_x + size_x * r * hapi.cos(angle + rotation)
        y = pivot_y + size_y * r * hapi.sin(angle + rotation)

        hapi.vertex((x, y))
    hapi.end_shape(fill=fill)


def smooth_star(hapi, x_coord, y_coord, size_x, size_y, n1=0.20, fill=False):
    """
    n1 between 0 and 1
    """
    smooth_star_options = {
        "n1": 0.20,
        "n2": 1.7,
        "n3": 1.7,
        "m": 5,
        "a": 1,
        "b": 1,
        "phi": 2,
    }
    smooth_star_options["n1"] = n1
    supershape(
        hapi,
        x_coord,
        y_coord,
        size_x,
        size_y,
        smooth_star_options,
        hapi._rotation,
        fill=fill,
    )


def oil_drop(hapi, x_coord, y_coord, size_x, size_y, n1=0.3, fill=False):
    """
    n1 between 0 and 1
    """
    oil_drop_options = {
        "n1": 0.3,
        "n2": 0.3,
        "n3": 0.3,
        "m": 1 / 6,
        "a": 1,
        "b": 1,
        "phi": 12,
    }
    oil_drop_options["n1"] = n1
    supershape(
        hapi,
        x_coord,
        y_coord,
        size_x,
        size_y,
        oil_drop_options,
        hapi._rotation,
        fill=fill,
    )


def flowing_star(hapi, x_coord, y_coord, size_x, size_y, n1=0.3, fill=False):
    flowing_star_options = {
        "n1": 0.3,
        "n2": 0.3,
        "n3": 0.3,
        "m": 7 / 6,
        "a": 1,
        "b": 1,
        "phi": 12,
    }
    flowing_star_options["n1"] = n1
    supershape(
        hapi,
        x_coord,
        y_coord,
        size_x,
        size_y,
        flowing_star_options,
        hapi._rotation,
        fill=fill,
    )


def cross_hair(hapi, coord):
    x = coord[0]
    y = coord[1]
    hapi.line(x, 0, x, hapi.HEIGHT)
    hapi.line(0, y, hapi.WIDTH, y)



import math
import colorsys


def constrain(val, start, end, realstart, realend):
    # mouseX 0 width 0 255
    # v = (mouseX / (end-start)) * (realend-realstart)
    # return realstart + v
    # if val < start, val = start
    # if val > end, val = end

    if val < start:
        return start
    if val > end:
        return end
    v = ((val - start) / (end - start)) * (realend - realstart)
    return realstart + v


def round_to_num(x, base=1):
    return base * round(x / base)


def distance(coord1, coord2):
    x1 = coord1[0]
    y1 = coord1[1]
    x2 = coord2[0]
    y2 = coord2[1]
    squared_sums = math.pow(x2 - x1, 2) + math.pow(y2 - y1, 2)
    d = math.pow(squared_sums, 0.5)
    return d


def rgb_to_hls(r, g, b):
    r, g, b = [x / 255.0 for x in (r, g, b)]
    h, l, s = colorsys.rgb_to_hls(r, g, b)
    return h, l, s


def hls_to_rgb(h, l, s):
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    r, g, b = [x * 255.0 for x in (r, g, b)]
    return r, g, b
